/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2013 Frederik Heber. All rights reserved.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * FitPartialChargesAction.cpp
 *
 *  Created on: Jul 03, 2013
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// needs to come before MemDebug due to placement new
#include <boost/archive/text_iarchive.hpp>

#include "CodePatterns/MemDebug.hpp"

#include "Atom/atom.hpp"
#include "CodePatterns/Log.hpp"
#include "Fragmentation/Exporters/ExportGraph_ToFiles.hpp"
#include "Fragmentation/Graph.hpp"
#include "World.hpp"

#include <boost/bimap.hpp>
#include <boost/bind.hpp>
#include <boost/filesystem.hpp>
#include <boost/foreach.hpp>
#include <algorithm>
#include <functional>
#include <iostream>
#include <string>

#include "Actions/PotentialAction/FitPartialChargesAction.hpp"

#include "Potentials/PartialNucleiChargeFitter.hpp"

#include "AtomIdSet.hpp"
#include "Descriptors/AtomIdDescriptor.hpp"
#include "Element/element.hpp"
#include "Element/periodentafel.hpp"
#include "Fragmentation/Homology/AtomFragmentsMap.hpp"
#include "Fragmentation/Homology/HomologyContainer.hpp"
#include "Fragmentation/Homology/HomologyGraph.hpp"
#include "Fragmentation/Summation/SetValues/SamplingGrid.hpp"
#include "FunctionApproximation/Extractors.hpp"
#include "Potentials/PartialNucleiChargeFitter.hpp"
#include "Potentials/Particles/ParticleFactory.hpp"
#include "Potentials/Particles/ParticleRegistry.hpp"
#include "Potentials/SerializablePotential.hpp"
#include "World.hpp"

using namespace MoleCuilder;

// and construct the stuff
#include "FitPartialChargesAction.def"
#include "Action_impl_pre.hpp"
/** =========== define the function ====================== */

static void enforceZeroTotalCharge(
    PartialNucleiChargeFitter::charges_t &_averaged_charges)
{
  double charge_sum = 0.;
  charge_sum = std::accumulate(_averaged_charges.begin(), _averaged_charges.end(), charge_sum);
  if (fabs(charge_sum) > MYEPSILON) {
    std::transform(
        _averaged_charges.begin(), _averaged_charges.end(),
        _averaged_charges.begin(),
        boost::bind(std::minus<double>(), _1, charge_sum/_averaged_charges.size()));
  }
  charge_sum = 0.;
  charge_sum = std::accumulate(_averaged_charges.begin(), _averaged_charges.end(), charge_sum);
  ASSERT( fabs(charge_sum) < MYEPSILON,
      "enforceZeroTotalCharge() - enforcing neutral net charge failed, "
      +toString(charge_sum)+" is the remaining net charge.");

  LOG(2, "DEBUG: final charges with net zero charge are " << _averaged_charges);
}

static size_t obtainAverageChargesOverFragments(
    PartialNucleiChargeFitter::charges_t &_average_charges,
    const std::pair<
      HomologyContainer::const_iterator,
      HomologyContainer::const_iterator> &_range,
    const double _radius
    )
{
  HomologyContainer::const_iterator iter = _range.first;
  if (!iter->second.containsGrids) {
    ELOG(1, "This HomologyGraph does not contain sampled grids.");
    return 0;
  }
  _average_charges.resize(iter->second.fragment.getCharges().size(), 0.);
  size_t NoFragments = 0;
  for (;
      iter != _range.second; ++iter, ++NoFragments) {
    if (!iter->second.containsGrids) {
      ELOG(2, "This HomologyGraph does not contain sampled grids,\ndid you forget to add '--store-grids 1' to AnalyseFragmentResults.");
      return 0;
    }
    const Fragment &fragment = iter->second.fragment;
  //  const double &energy = iter->second.energy;
  //  const SamplingGrid &charge = iter->second.charge_distribution;
    const SamplingGrid &potential = iter->second.potential_distribution;
    if ((potential.level == 0)
        || ((potential.begin[0] == potential.end[0])
            && (potential.begin[1] == potential.end[1])
            && (potential.begin[2] == potential.end[2]))) {
      ELOG(1, "Sampled grid contains grid made of zero points.");
      return 0;
    }

    // then we extract positions from fragment
    PartialNucleiChargeFitter::positions_t positions;
    Fragment::positions_t fragmentpositions = fragment.getPositions();
    positions.reserve(fragmentpositions.size());
    BOOST_FOREACH( Fragment::position_t pos, fragmentpositions) {
      positions.push_back( Vector(pos[0], pos[1], pos[2]) );
    }
    PartialNucleiChargeFitter fitter(potential, positions, _radius);
    fitter();
    PartialNucleiChargeFitter::charges_t return_charges = fitter.getSolutionAsCharges_t();
    LOG(2, "DEBUG: fitted charges are " << return_charges);
    std::transform(
        return_charges.begin(), return_charges.end(),
        _average_charges.begin(),
        _average_charges.begin(),
        std::plus<PartialNucleiChargeFitter::charge_t>());
  }
  if (NoFragments != 0)
    std::transform(_average_charges.begin(), _average_charges.end(),
        _average_charges.begin(),
        std::bind1st(std::multiplies<PartialNucleiChargeFitter::charge_t>(),1./(double)NoFragments)
    );
  LOG(2, "DEBUG: final averaged charges are " << _average_charges);

  return NoFragments;
}

inline SerializablePotential::ParticleTypes_t
getParticleTypesForAtomIdSet(const AtomIdSet &_atoms)
{
  SerializablePotential::ParticleTypes_t particletypes;
  particletypes.resize(_atoms.size());
  std::transform(
      _atoms.begin(), _atoms.end(),
      particletypes.begin(),
      boost::bind(&atom::getElementNo, _1));
  return particletypes;
}

ActionState::ptr PotentialFitPartialChargesAction::performCall()
{
  // check for selected atoms
  const World &world = const_cast<const World &>(World::getInstance());
  if (world.beginAtomSelection() == world.endAtomSelection()) {
    STATUS("There are no atoms selected for fitting partial charges to.");
    return Action::failure;
  }

  /// obtain possible fragments to each selected atom
  const AtomFragmentsMap &atomfragments = AtomFragmentsMap::getConstInstance();
  if (!atomfragments.checkCompleteness()) {
    STATUS("AtomFragmentsMap failed internal consistency check, missing forcekeysets?");
    return Action::failure;
  }
  std::set<KeySet> fragments;
  const AtomFragmentsMap::AtomFragmentsMap_t &atommap = atomfragments.getMap();
  for (World::AtomSelectionConstIterator iter = world.beginAtomSelection();
      iter != world.endAtomSelection(); ++iter) {
    const AtomFragmentsMap::AtomFragmentsMap_t::const_iterator atomiter =
        atommap.find(iter->first);
    if (iter->second->getElementNo() != 1) {
      if (atomiter == atommap.end()) {
        ELOG(2, "There are no fragments associated to " << iter->first << ".");
        continue;
      }
      const AtomFragmentsMap::keysets_t &keysets = atomiter->second;
      LOG(2, "DEBUG: atom " << iter->first << " has " << keysets.size() << " fragments.");
      fragments.insert( keysets.begin(), keysets.end() );
    } else {
      LOG(3, "DEBUG: Skipping atom " << iter->first << " as it's hydrogen.");
    }
  }

  // reduce given fragments to homologous graphs to avoid multiple fittings
  typedef std::map<KeySet, HomologyGraph> KeysetsToGraph_t;
  KeysetsToGraph_t keyset_graphs;
  typedef std::map<HomologyGraph, PartialNucleiChargeFitter::charges_t> GraphFittedChargeMap_t;
  GraphFittedChargeMap_t fittedcharges_per_fragment;
  HomologyContainer &homologies = World::getInstance().getHomologies();
  for (std::set<KeySet>::const_iterator fragmentiter = fragments.begin();
      fragmentiter != fragments.end(); ++fragmentiter) {
    const KeySet &keyset = *fragmentiter;
    const AtomFragmentsMap::indices_t &forceindices = atomfragments.getFullKeyset(keyset);
    ASSERT( !forceindices.empty(),
        "PotentialFitPartialChargesAction::performCall() - force keyset to "
        +toString(keyset)+" is empty.");
    KeySet forcekeyset;
    forcekeyset.insert(forceindices.begin(), forceindices.end());
    forcekeyset.erase(-1);
    const HomologyGraph graph(forcekeyset);
    LOG(2, "DEBUG: Associating keyset " << forcekeyset << " with graph " << graph);
    keyset_graphs.insert( std::make_pair(keyset, graph) );
    fittedcharges_per_fragment.insert( std::make_pair(graph, PartialNucleiChargeFitter::charges_t()) );
  }

  /// then go through all fragments and get partial charges for each
  for (GraphFittedChargeMap_t::iterator graphiter = fittedcharges_per_fragment.begin();
      graphiter != fittedcharges_per_fragment.end(); ++graphiter) {
    const HomologyGraph &graph = graphiter->first;
    LOG(2, "DEBUG: Now fitting charges to graph " << graph);
    const HomologyContainer::range_t range = homologies.getHomologousGraphs(graph);
    if (range.first == range.second) {
      STATUS("HomologyContainer does not contain specified fragment.");
      return Action::failure;
    }

    // fit and average partial charges over all homologous fragments
    PartialNucleiChargeFitter::charges_t &averaged_charges = graphiter->second;
    const size_t NoFragments =
        obtainAverageChargesOverFragments(averaged_charges, range, params.radius.get());
    if ((NoFragments == 0) && (range.first != range.second)) {
      STATUS("Fitting for fragment "+toString(*graphiter)+" failed.");
      return Action::failure;
    }

    // make sum of charges zero if desired
    if (params.enforceZeroCharge.get())
      enforceZeroTotalCharge(averaged_charges);

    // output status info fitted charges
    LOG(2, "DEBUG: For fragment " << *graphiter << " we have fitted the following charges "
        << averaged_charges << ", averaged over " << NoFragments << " fragments.");
  }

  /// obtain average charge for each atom the fitted charges over all its fragments
  typedef std::map<atomId_t, double> fitted_charges_t;
  fitted_charges_t fitted_charges;
  for (World::AtomSelectionConstIterator atomiter = world.beginAtomSelection();
      atomiter != world.endAtomSelection(); ++atomiter) {
    const atom * walker = atomiter->second;
    if (walker->getElementNo() == 1) {
      // it's hydrogen, check its bonding and use its bond partner instead to request
      // keysets
      const BondList &ListOfBonds = walker->getListOfBonds();
      if ( ListOfBonds.size() != 1) {
        ELOG(1, "Solitary hydrogen in atom " << atomiter->first << " detected, skipping.");
        continue;
      }
      walker = (*ListOfBonds.begin())->GetOtherAtom(walker);
    }
    const atomId_t walkerid = walker->getId();
    const AtomFragmentsMap::AtomFragmentsMap_t::const_iterator keysetsiter =
        atommap.find(walkerid);
    ASSERT(keysetsiter != atommap.end(),
        "PotentialFitPartialChargesAction::performCall() - we checked already that "
        +toString(walkerid)+" should be present!");
    const AtomFragmentsMap::keysets_t & keysets = keysetsiter->second;

    double average_charge = 0.;
    size_t NoFragments = 0;
    // go over all fragments associated to this atom
    for (AtomFragmentsMap::keysets_t::const_iterator keysetsiter = keysets.begin();
        keysetsiter != keysets.end(); ++keysetsiter) {
      const KeySet &keyset = *keysetsiter;

      const AtomFragmentsMap::indices_t &forcekeyset = atomfragments.getFullKeyset(keyset);
      ASSERT( !forcekeyset.empty(),
          "PotentialFitPartialChargesAction::performCall() - force keyset to "
          +toString(keyset)+" is empty.");

      // find the associated charge in the charge vector
      const HomologyGraph &graph = keyset_graphs[keyset];
      const GraphFittedChargeMap_t::const_iterator chargesiter =
          fittedcharges_per_fragment.find(graph);
      ASSERT(chargesiter != fittedcharges_per_fragment.end(),
          "PotentialFitPartialChargesAction::performCall() - no charge to "+toString(keyset)
          +" any longer present in fittedcharges_per_fragment?");
      const PartialNucleiChargeFitter::charges_t &charges = chargesiter->second;
      ASSERT( charges.size() == forcekeyset.size(),
          "PotentialFitPartialChargesAction::performCall() - charges "
          +toString(charges.size())+" and keyset "+toString(forcekeyset.size())
          +" do not have the same length?");
      PartialNucleiChargeFitter::charges_t::const_iterator chargeiter =
          charges.begin();
      const AtomFragmentsMap::indices_t::const_iterator forcekeysetiter =
          std::find(forcekeyset.begin(), forcekeyset.end(), atomiter->first);
      ASSERT( forcekeysetiter != forcekeyset.end(),
          "PotentialFitPartialChargesAction::performCall() - atom "
          +toString(atomiter->first)+" not contained in force keyset "+toString(forcekeyset));
      std::advance(chargeiter, std::distance(forcekeyset.begin(), forcekeysetiter));

      // and add onto charge sum
      const double & charge_in_fragment = *chargeiter;
      average_charge += charge_in_fragment;
      ++NoFragments;
    }
    // average to obtain final partial charge for this atom
    average_charge *= 1./(double)NoFragments;
    fitted_charges.insert( std::make_pair(atomiter->first, average_charge) );

    LOG(2, "DEBUG: For atom " << atomiter->first << " we have an average charge of "
        << average_charge);
  }

  /// place all fitted charges into ParticleRegistry
  const ParticleFactory &factory =
      const_cast<const ParticleFactory&>(ParticleFactory::getInstance());
  const periodentafel &periode = *World::getInstance().getPeriode();
  for (fitted_charges_t::const_iterator chargeiter = fitted_charges.begin();
      chargeiter != fitted_charges.end(); ++chargeiter) {
    const atom * const walker = World::getInstance().getAtom(AtomById(chargeiter->first));
    ASSERT( walker != NULL,
        "PotentialFitPartialChargesAction::performCall() - atom "
        +toString(chargeiter->first)+" not present in the World?");
    const double &charge = chargeiter->second;
    const atomicNumber_t elementno = walker->getElementNo();
    const std::string name = Particle::findFreeName(periode, elementno);
    LOG(1, "INFO: Adding particle " << name << " for atom "
        << *walker << " with element " << elementno << ", charge " << charge);
    factory.createInstance(name, elementno, charge);
  }

  return Action::success;
}

ActionState::ptr PotentialFitPartialChargesAction::performUndo(ActionState::ptr _state) {
  return Action::success;
}

ActionState::ptr PotentialFitPartialChargesAction::performRedo(ActionState::ptr _state){
  return Action::success;
}

bool PotentialFitPartialChargesAction::canUndo() {
  return false;
}

bool PotentialFitPartialChargesAction::shouldUndo() {
  return false;
}
/** =========== end of function ====================== */
